@@ -13,8 +13,8 @@
 
 luatexbase.provides_module {
   name        = 'luatexko',
-  date        = '2020/02/17',
-  version     = '2.6',
+  date        = '2020/07/27',
+  version     = '2.9',
   description = 'typesetting Korean with LuaTeX',
   author      = 'Dohyun Kim, Soojin Nam',
   license     = 'LPPL v1.3+',
@@ -36,11 +36,9 @@ local insert_before   = node.insert_before
 local nodecopy        = node.copy
 local nodecount       = node.count
 local nodefree        = node.free
-local nodeid          = node.id
 local nodenew         = node.new
 local noderemove      = node.remove
 local nodeslide       = node.slide
-local nodesubtype     = node.subtype
 local nodewrite       = node.write
 local rangedimensions = node.rangedimensions
 local setglue         = node.setglue
@@ -75,7 +73,6 @@ local callback_descriptions = luatexbase.callback_descriptions
 local create_callback       = luatexbase.create_callback
 local module_warning        = luatexbase.module_warning
 local new_user_whatsit      = luatexbase.new_user_whatsit
-local new_user_whatsit_id   = luatexbase.new_user_whatsit_id
 local registernumber        = luatexbase.registernumber
 local remove_from_callback  = luatexbase.remove_from_callback
 
@@ -83,19 +80,18 @@ local function warning (...)
   return module_warning("luatexko", stringformat(...))
 end
 
-local dirid     = nodeid"dir"
-local discid    = nodeid"disc"
-local glueid    = nodeid"glue"
-local glyphid   = nodeid"glyph"
-local hlistid   = nodeid"hlist"
-local kernid    = nodeid"kern"
-local mathid    = nodeid"math"
-local penaltyid = nodeid"penalty"
-local ruleid    = nodeid"rule"
-local vlistid   = nodeid"vlist"
-local whatsitid = nodeid"whatsit"
-local literal_whatsit = nodesubtype"pdf_literal"
-local user_whatsit    = nodesubtype"user_defined"
+local dirid     = node.id"dir"
+local discid    = node.id"disc"
+local glueid    = node.id"glue"
+local glyphid   = node.id"glyph"
+local hlistid   = node.id"hlist"
+local kernid    = node.id"kern"
+local mathid    = node.id"math"
+local penaltyid = node.id"penalty"
+local ruleid    = node.id"rule"
+local vlistid   = node.id"vlist"
+local whatsitid = node.id"whatsit"
+local literal_whatsit = node.subtype"pdf_literal"
 local directmode = 2
 local fontkern   = 0
 local userkern   = 1
@@ -112,6 +108,7 @@ local fallbackfontattr = attributes.luatexkofallbackfontattr
 local autojosaattr     = attributes.luatexkoautojosaattr
 local classicattr      = attributes.luatexkoclassicattr
 local dotemphattr      = attributes.luatexkodotemphattr
+local ulineattr        = attributes.luatexkoulineattr
 local rubyattr         = attributes.luatexkorubyattr
 local hangulbyhangulattr = attributes.luatexkohangulbyhangulattr
 local hanjabyhanjaattr   = attributes.luatexkohanjabyhanjaattr
@@ -467,13 +464,12 @@ luatexko.forcehangulchars = force_hangul
 
 local forcehf_f, forcehf_id = new_user_whatsit("forcehf","luatexko")
 
-local function update_force_hangul (value)
+function luatexko.updateforcehangul (value)
   local what = forcehf_f()
   what.type  = lua_value -- function
   what.value = value
   nodewrite(what)
 end
-luatexko.updateforcehangul = update_force_hangul
 
 local function hangul_space_skip (curr, newfont)
   if curr.lang ~= nohyphen and curr.font ~= newfont then
@@ -519,7 +515,6 @@ local function process_fonts (head)
              is_not_harf(curr.font) and
              option_in_font(curr.font, "script") == "hang" then
             luatexko.activate("reorderTM") -- activate reorderTM here
-            active_processes.reorderTM = true
           end
 
         else
@@ -558,10 +553,9 @@ local function process_fonts (head)
       curr.replace = process_fonts(curr.replace)
     elseif id == mathid then
       curr = end_of_math(curr)
-    elseif id == whatsitid
-      and curr.subtype == user_whatsit
-      and curr.user_id == forcehf_id
-      and curr.type == lua_value then
+    elseif id      == whatsitid  and
+      curr.user_id == forcehf_id and
+      curr.type    == lua_value  then
 
       local value = curr.value
       if type(value) == "function" then
@@ -903,9 +897,7 @@ local function process_interhangul (head, par)
       if c and not is_unicode_var_sel(c) then
         head, pc = do_interhangul_option(head, curr, pc, c, curr.font, par)
 
-        if is_chosong(c) then
-          pc = 0
-        elseif is_jungsong(c) or is_jongsong(c) or hangul_tonemark[c] then
+        if is_jungsong(c) or is_jongsong(c) or hangul_tonemark[c] then
           pc = 1
         end
       end
@@ -1125,10 +1117,8 @@ local josa_code = setmetatable({
   end
 end })
 
-local ignore_parens = false
-
-local function prevjosacode (n, parenlevel)
-  local parenlevel, josacode = parenlevel or 0
+local function prevjosacode (n, parenlevel, ignore_parens)
+  local josacode
   while n do
     local id = n.id
     if id == glyphid then
@@ -1144,7 +1134,7 @@ local function prevjosacode (n, parenlevel)
     elseif id == hlistid or id == vlistid then
       local list = n.list
       if list then
-        josacode, parenlevel = prevjosacode(nodeslide(list), parenlevel)
+        josacode, parenlevel = prevjosacode(nodeslide(list), parenlevel, ignore_parens)
         if josacode then break end
       end
     end
@@ -1170,8 +1160,7 @@ local function process_josa (head)
         end
         local t = josa_table[cc]
         if t then
-          ignore_parens = autojosaattr > 0
-          cc = t[prevjosacode(getprev(curr)) or 3]
+          cc = t[prevjosacode(getprev(curr), 0, autojosaattr > 0) or 3]
           if cc then
             curr.char = cc
           else
@@ -1215,11 +1204,22 @@ end
 local dotemphbox = {}
 luatexko.dotemphbox = dotemphbox
 
-local function process_dotemph (head, tofree)
-  local curr, outer, tofree = head, not tofree, tofree or {}
+local dotemph_f, dotemph_id = new_user_whatsit("dotemph","luatexko")
+
+function luatexko.dotemphboundary (i)
+  local what = dotemph_f()
+  what.type  = lua_number
+  what.value = i
+  nodewrite(what)
+end
+
+local function process_dotemph (head)
+  local curr = head
   while curr do
-    local id = curr.id
-    if id == glyphid then
+    if curr.list then
+      curr.list = process_dotemph(curr.list)
+
+    elseif curr.id == glyphid then
       local dotattr = has_attribute(curr, dotemphattr)
       if dotattr then
         local c = my_node_props(curr).unicode or curr.char
@@ -1236,7 +1236,13 @@ local function process_dotemph (head, tofree)
 
           local currwd = curr.width
           if currwd >= font_options.en_size[curr.font] then
-            local box = nodecopy(dotemphbox[dotattr])
+            local box = nodecopy(dotemphbox[dotattr]).list
+            -- bypass unwanted nodes injected by some other packages
+            while box.id ~= hlistid do
+              warning[[\dotemph should be an hbox]]
+              box = getnext(box)
+            end
+
             local shift = (currwd - box.width)/2
             if shift ~= 0 then
               local list = box.list
@@ -1250,28 +1256,26 @@ local function process_dotemph (head, tofree)
 
             box.width = 0
             head = insert_before(head, curr, box)
-            tofree[dotattr] = true
           end
         end
-        unset_attribute(curr, dotemphattr)
-      end
-    elseif id == hlistid then
-      local list = curr.list
-      if list then
-        curr.list, tofree = process_dotemph(list, tofree)
       end
+
+    elseif curr.id == whatsitid  and
+      curr.user_id == dotemph_id and
+      curr.type    == lua_number then
+
+      local val = curr.value
+      nodefree(dotemphbox[val])
+      dotemphbox[val] = nil
     end
     curr = getnext(curr)
   end
-  if outer then
-    for k in pairs(tofree) do nodefree(dotemphbox[k]) end
-  end
-  return head, tofree
+  return head
 end
 
 -- uline
 
-local function get_strike_out_down (box)
+function luatexko.get_strike_out_down (box)
   local c, f = hbox_char_font(box, true, true) -- ignore blocking nodes
   if c and f then
     local down
@@ -1291,14 +1295,16 @@ local function get_strike_out_down (box)
   end
   return -texsp"0.5ex"
 end
-luatexko.get_strike_out_down = get_strike_out_down
 
 local uline_f, uline_id = new_user_whatsit("uline","luatexko")
-local no_uline_id = new_user_whatsit_id("no_uline","luatexko")
 
-local function ulboundary (i, n, subtype)
+function luatexko.ulboundary (i, n, subtype)
   local what = uline_f()
   if n then
+    while n.id ~= ruleid and n.id ~= hlistid and n.id ~= vlistid do
+      warning[[\markoverwith should be a rule or a box]]
+      n = getnext(n)
+    end
     what.type  = lua_value -- table
     what.value = { i, nodecopy(n), subtype }
   else
@@ -1307,7 +1313,6 @@ local function ulboundary (i, n, subtype)
   end
   nodewrite(what)
 end
-luatexko.ulboundary = ulboundary
 
 local white_nodes = {
   [glueid]    = true,
@@ -1316,7 +1321,7 @@ local white_nodes = {
   [whatsitid] = true,
 }
 
-function skip_white_nodes (n, ltr)
+local function skip_white_nodes (n, ltr)
   local nextnode = ltr and getnext or getprev
   while n do
     if not white_nodes[n.id] then break end
@@ -1328,66 +1333,78 @@ end
 local function draw_uline (head, curr, parent, t, final)
   local start, list, subtype = t.start or head, t.list, t.subtype
   start = skip_white_nodes(start, true)
-  if final then
+  if final and start then
     nodeslide(start) -- to get correct getprev.
   end
   curr  = skip_white_nodes(curr)
-  curr  = getnext(curr) or curr
-  local len = parent and rangedimensions(parent, start, curr)
-                     or  dimensions(start, curr) -- it works?!
-  if len and len ~= 0 then
-    local g = nodenew(glueid)
-    setglue(g, len)
-    g.subtype = subtype
-    g.leader = final and list or nodecopy(list)
-    local k = nodenew(kernid)
-    k.kern = -len
-    head = insert_before(head, start, g)
-    head = insert_before(head, start, k)
+  if start and curr then
+    curr = getnext(curr) or curr
+    local len = parent and rangedimensions(parent, start, curr)
+                       or  dimensions(start, curr)
+    if len and len ~= 0 then
+      local g = nodenew(glueid)
+      setglue(g, len)
+      g.subtype = subtype
+      g.leader  = final and list or nodecopy(list)
+      local k = nodenew(kernid)
+      k.kern = -len
+      head = insert_before(head, start, g)
+      head = insert_before(head, start, k)
+    end
   end
   return head
 end
 
-local function process_uline (head, parent, items, level)
-  local curr, items, level = head, items or {}, level or 0
+local ulitems = {}
+
+local function process_uline (head, parent, level)
+  local curr, level, attr = head, level or 0
   while curr do
-    local id = curr.id
-    if id == whatsitid
-      and curr.subtype == user_whatsit
-      and curr.user_id == uline_id then
+    if curr.list then
+      curr.list = process_uline(curr.list, curr, level+1)
+
+    elseif curr.id == whatsitid and curr.user_id == uline_id then
 
       local value = curr.value
       if curr.type == lua_value then
         local count, list, subtype = tableunpack(value)
-        items[count] = {
-          start   = curr,
+        ulitems[count] = {
           list    = list,
           subtype = subtype,
           level   = level,
         }
-      elseif items[value] then
-        head = draw_uline(head, curr, parent, items[value], true)
-        items[value] = nil
+      else
+        local item = ulitems[value]
+        if item then
+          head = draw_uline(head, curr, parent, item, true)
+          ulitems[value] = nil
+          attr = nil
+        end
       end
 
-      curr.user_id = no_uline_id -- avoid multiple run
-    elseif id == hlistid then
-      local list = curr.list
-      if list then
-        curr.list, items = process_uline(list, curr, items, level+1)
+    else
+      local currattr = has_attribute(curr, ulineattr)
+      if currattr then
+        local item = ulitems[currattr]
+        if item and not item.start then
+          item.start = curr
+          attr = currattr
+        end
       end
+
     end
     curr = getnext(curr)
   end
 
-  curr = nodeslide(head)
-  for i, t in pairs(items) do
-    if level == t.level then
-      head = draw_uline(head, curr, parent, t)
-      t.start = nil
+  if attr then
+    local item = ulitems[attr]
+    if item and item.level == level then
+      head = draw_uline(head, nodeslide(head), parent, item)
+      item.start = nil
     end
   end
-  return head, items
+
+  return head
 end
 
 -- ruby
@@ -1395,7 +1412,7 @@ end
 local rubybox = {}
 luatexko.rubybox = rubybox
 
-local function getrubystretchfactor (box)
+function luatexko.getrubystretchfactor (box)
   local _, fid = hbox_char_font(box, true, true)
   local str = font_options.intercharstretch[fid]
   if str then
@@ -1403,7 +1420,6 @@ local function getrubystretchfactor (box)
     set_macro("luatexkostretchfactor", stringformat("%.4f", str/em/2))
   end
 end
-luatexko.getrubystretchfactor = getrubystretchfactor
 
 local function process_ruby_pre_linebreak (head)
   local curr = head
@@ -1463,13 +1479,11 @@ local function process_ruby_post_linebreak (head)
           head = insert_before(head, curr, ruby)
         end
         ruby_t = nil
-        unset_attribute(curr, rubyattr)
-      else
-        local list = curr.list
-        if list then
-          curr.list = process_ruby_post_linebreak(list)
-        end
+      elseif curr.list then
+        curr.list = process_ruby_post_linebreak(curr.list)
       end
+    elseif curr.list then
+      curr.list = process_ruby_post_linebreak(curr.list)
     end
     curr = getnext(curr)
   end
@@ -1736,7 +1750,7 @@ local function process_vertical_font (fontdata)
   local goffset = font_opt_dim(fontdata, "charraise") or
                   (parameters.x_height or quad/2)/2
   -- declare shift amount of horizontal box inside vertical env.
-  fontdata.horizboxmoveleftamount = quad/2-goffset
+  fontdata.horizboxmoverightamount = goffset
 
   for i,v in pairs(fontdata.characters) do
     local voff = goffset - (v.width or 0)/2
@@ -1802,21 +1816,20 @@ local function process_vertical_font (fontdata)
   end
 end
 
-local function get_horizbox_moveleft ()
+function luatexko.gethorizboxmoveright ()
   for _, v in ipairs{ fontcurrent(),
                       texattribute.luatexkohangulfontattr,
                       texattribute.luatexkohanjafontattr,
                       texattribute.luatexkofallbackfontattr } do
     if v and v > 0 then
-      local amount = get_font_data(v).horizboxmoveleftamount
+      local amount = get_font_data(v).horizboxmoverightamount
       if amount then
-        set_macro("luatexkohorizboxmoveleft", texsp(amount).."sp")
+        set_macro("luatexkohorizboxmoveright", texsp(amount).."sp")
         break
       end
     end
   end
 end
-luatexko.gethorizboxmoveleft = get_horizbox_moveleft
 
 -- charraise
 
@@ -1890,10 +1903,13 @@ local function process_fake_slant_font (fontdata)
   local fsl = fontdata.slant
   if fsl and fsl > 0 then
     fsl = fsl/1000
-    local params = fontdata.parameters or {}
-    params.slant = (params.slant or 0) + fsl*65536 -- slant per point
-
     local hb = is_harf(fontdata)
+
+    local params = fontdata.parameters or {}
+    if not hb then -- hb done by luaotfload
+      params.slant = (params.slant or 0) + fsl*65536 -- slant per point
+    end
+
     local scale  = hb and hb.scale or params.factor or 655.36
 
     local shared = fontdata.shared or {}
@@ -1933,6 +1949,13 @@ create_callback("luatexko_hpack_first",         "data", pass_fun)
 create_callback("luatexko_prelinebreak_first",  "data", pass_fun)
 create_callback("luatexko_hpack_second",        "data", pass_fun)
 create_callback("luatexko_prelinebreak_second", "data", pass_fun)
+create_callback("luatexko_do_atbegshi",         "data", pass_fun)
+
+function luatexko.process_atbegshi (box)
+  if box and box.list then
+    box.list = call_callback("luatexko_do_atbegshi", box.list)
+  end
+end
 
 add_to_callback("hpack_filter", function(h)
   h = process_fonts(h)
@@ -1975,14 +1998,43 @@ local font_opt_procs = {
   },
 }
 
+-- luaotfload expansion feature, committed at 2020.06.08
+luatexko.plug_expansion_loaded = false
+for _,v in next, fonts.handlers.otf.features.manipulators.plug do
+  if v.name == "expansion" then
+    luatexko.plug_expansion_loaded = true
+    break
+  end
+end
+
 local font_opt_procs_single = {
-  expansion = function()
+  expansion = function(fontdata)
+    if not luatexko.plug_expansion_loaded and is_harf(fontdata) then
+      local val   = option_in_font(fontdata, "expansion")
+      local setup = fonts.expansions.setups[val] or {}
+      fontdata.stretch = fontdata.stretch or (setup.stretch or 2)*10
+      fontdata.shrink  = fontdata.shrink  or (setup.shrink  or 2)*10
+      fontdata.step    = fontdata.step    or (setup.step    or .5)*10
+    end
     if tex.adjustspacing == 0 then
       texset("global", "adjustspacing", 2)
     end
   end,
 
   protrusion = function(fontdata)
+    if is_harf(fontdata) then -- for now, no support in harf mode
+      local val   = option_in_font(fontdata, "protrusion")
+      local setup = fonts.protrusions.setups[val] or {}
+      local quad  = fontdata.parameters.quad
+      for i, v in pairs(setup) do
+        local chr = fontdata.characters[i]
+        if chr then
+          local wdq = chr.width/quad
+          chr.left_protruding  = chr.left_protruding  or wdq*v[1]*1000
+          chr.right_protruding = chr.right_protruding or wdq*v[2]*1000
+        end
+      end
+    end
     if tex.protrudechars == 0 then
       texset("global", "protrudechars", 2)
     end
@@ -2043,17 +2095,14 @@ add_to_callback("luaotfload.patch_font_unsafe", process_patch_font,
 
 local auxiliary_procs = {
   dotemph = {
-    hpack_filter = process_dotemph,
-    vpack_filter = process_dotemph,
+    luatexko_do_atbegshi = process_dotemph,
   },
   uline   = {
-    hpack_filter = process_uline,
-    vpack_filter = process_uline,
+    luatexko_do_atbegshi = process_uline,
   },
   ruby    = {
     pre_linebreak_filter = process_ruby_pre_linebreak,
-    hpack_filter         = process_ruby_post_linebreak,
-    vpack_filter         = process_ruby_post_linebreak,
+    luatexko_do_atbegshi = process_ruby_post_linebreak,
   },
   autojosa = {
     luatexko_hpack_first        = process_josa,
@@ -2065,26 +2114,12 @@ local auxiliary_procs = {
   },
 }
 
-local function activate (name)
+function luatexko.activate (name)
   for cbnam, cbfun in pairs( auxiliary_procs[name] ) do
-    local fun
-    if cbnam == "hpack_filter" then
-      fun = function(h, gc)
-        if gc == "align_set" then
-          h = cbfun(h)
-        end
-        return h
-      end
-    else
-      fun = function(h)
-        h = cbfun(h)
-        return h
-      end
-    end
-    add_to_callback(cbnam, fun, "luatexko."..cbnam.."."..name)
+    add_to_callback(cbnam, cbfun, "luatexko."..cbnam.."."..name)
   end
+  active_processes[name] = true
 end
-luatexko.activate = activate
 
 add_to_callback ("hyphenate",
 function(head)
@@ -2114,12 +2149,12 @@ end,
 
 -- aux functions
 
-local function deactivate_all (str)
+function luatexko.deactivateall (str)
   luatexko.deactivated = {}
   for _, name in ipairs{ "hpack_filter",
                          "pre_linebreak_filter",
-                         "vpack_filter",
                          "hyphenate",
+                         "luatexko_do_atbegshi", -- might not work properly
                          "luaotfload.patch_font_unsafe", -- added for harf
                          "luaotfload.patch_font" } do
     local t = {}
@@ -2132,9 +2167,8 @@ local function deactivate_all (str)
     luatexko.deactivated[name] = t
   end
 end
-luatexko.deactivateall = deactivate_all
 
-local function reactivate_all ()
+function luatexko.reactivateall ()
   for name, v in pairs(luatexko.deactivated or {}) do
     for _, vv in ipairs(v) do
       add_to_callback(name, tableunpack(vv))
@@ -2142,10 +2176,8 @@ local function reactivate_all ()
   end
   luatexko.deactivated = nil
 end
-luatexko.reactivateall = reactivate_all
 
-local function current_has_hangul_chars (cnt)
+function luatexko.currenthashangulchars (cnt)
   texcount[cnt] = char_in_font(fontcurrent(), 0xAC00) and 1 or 0
 end
-luatexko.currenthashangulchars = current_has_hangul_chars
 
