@@ -1,6 +1,6 @@
-# $Id: TLDownload.pm 32912 2014-02-07 23:49:53Z karl $
+# $Id: TLDownload.pm 54993 2020-05-03 21:57:54Z karl $
 # TeXLive::TLDownload.pm - module for abstracting the download modes
-# Copyright 2009-2014 Norbert Preining
+# Copyright 2009-2020 Norbert Preining
 # This file is licensed under the GNU General Public License version 2
 # or any later version.
 
@@ -8,9 +8,8 @@ package TeXLive::TLDownload;
 
 use TeXLive::TLUtils;
 use TeXLive::TLConfig;
-use File::Temp qw/tempfile/;
 
-my $svnrev = '$Revision: 32912 $';
+my $svnrev = '$Revision: 54993 $';
 my $_modulerevision;
 if ($svnrev =~ m/: ([0-9]+) /) {
   $_modulerevision = $1;
@@ -21,8 +20,6 @@ sub module_revision {
   return $_modulerevision;
 }
 
-my $MAX_ERRORCOUNT = 5;
-
 # since Net::HTTP and Net::FTP are shipped by the same packages
 # we only test for Net::HTTP, if that fails, let us know ;-)
 our $net_lib_avail = 0;
@@ -42,6 +39,18 @@ sub new
 {
   my $class = shift;
   my $self = {};
+  $self->{'initcount'} = 0;
+  bless $self, $class;
+  $self->reinit();
+  return $self;
+}
+
+
+
+
+sub reinit
+{
+  my $self = shift;
   my $ua = LWP::UserAgent->new(
     agent => "texlive/lwp",
     # use LWP::ConnCache, and keep 1 connection open
@@ -52,13 +61,8 @@ sub new
   $self->{'ua'} = $ua;
   $self->{'enabled'} = 1;
   $self->{'errorcount'} = 0;
-  bless $self, $class;
-  return $self;
+  $self->{'initcount'} += 1;
 }
-
-
-
-
 sub enabled
 {
   my $self = shift;
@@ -81,6 +85,11 @@ sub disable
   my $self = shift;
   $self->{'enabled'} = 0;
 }
+sub initcount
+{
+  my $self = shift;
+  return $self->{'initcount'};
+}
 sub errorcount
 {
   my $self = shift;
@@ -111,7 +120,7 @@ sub get_file {
   my ($self,$url,$out,$size) = @_;
   #
   # automatically disable if error count is getting too big
-  if ($self->errorcount > $MAX_ERRORCOUNT) {
+  if ($self->errorcount > $TeXLive::TLConfig::MaxLWPErrors) {
     $self->disable;
   }
   # return if disabled
@@ -120,7 +129,7 @@ sub get_file {
   my $realout = $out;
   my ($outfh, $outfn);
   if ($out eq "|") {
-    ($outfh, $outfn) = tempfile();
+    ($outfh, $outfn) = tl_tmpfile();
     $realout = $outfn;
   }
   my $response = $self->{'ua'}->get($url, ':content_file' => $realout);
@@ -134,7 +143,7 @@ sub get_file {
       return $outfh;
     }
   } else {
-    tlwarn("TLDownload::get_file: response error: "
+    debug("TLDownload::get_file: response error: "
             . $response->status_line . " (for $url)\n");
     $self->incr_errorcount;
     return;
